#ifndef ATMEGA88PA_H_INCLUDED
#define ATMEGA88PA_H_INCLUDED

/* System Clock */
// Clock Calibration
#define CLOCK_CALIBRATION_VALUE(value) \
    OSCCAL = (value)
#define CLOCK_INCR_CALIBRATION_VALUE(offset) \
    OSCCAL += (offset)
#define CLOCK_DECR_CALIBRATION_VALUE(offset) \
    OSCCAL -= (offset)

/* Timer/Counter0 */
// Modes of Operation
#define TIMER0_NORMAL_MODE \
    TCCR0B &= ~_BV(WGM02); \
    TCCR0A = (TCCR0A & ~_BV(WGM01)) & ~_BV(WGM00)
#define TIMER0_PHASE_CORRECT_PWM_FULL_RANGE_MODE \
    TCCR0B &= ~_BV(WGM02); \
    TCCR0A = (TCCR0A & ~_BV(WGM01)) | _BV(WGM00)
#define TIMER0_CLEAR_TIMER_ON_COMPARE_MATCH_MODE \
    TCCR0B &= ~_BV(WGM02); \
    TCCR0A = (TCCR0A | _BV(WGM01)) & ~_BV(WGM00)
#define TIMER0_FAST_PWM_FULL_RANGE_MODE \
    TCCR0B &= ~_BV(WGM02); \
    TCCR0A = (TCCR0A | _BV(WGM01)) | _BV(WGM00)
#define TIMER0_PHASE_CORRECT_PWM_CONTROL_RANGE_MODE \
    TCCR0B |= _BV(WGM02); \
    TCCR0A = (TCCR0A & ~_BV(WGM01)) | _BV(WGM00)
#define TIMER0_FAST_PWM_CONTROL_RANGE_MODE \
    TCCR0B |= _BV(WGM02); \
    TCCR0A = (TCCR0A | _BV(WGM01)) | _BV(WGM00)
// Timer/Counter0 Compare Output Mode, Fast PWM Mode
#define TIMER0_PWMA_NORMAL_PORT_OPERATION \
    TCCR0A = (TCCR0A & ~_BV(COM0A1)) & ~_BV(COM0A0)
#define TIMER0_PWMA_CLEAR_ON_COMPARE_MATCH \
    TCCR0A = (TCCR0A | _BV(COM0A1)) & ~_BV(COM0A0)
#define TIMER0_PWMA_SET_ON_COMPARE_MATCH \
    TCCR0A = (TCCR0A | _BV(COM0A1)) | _BV(COM0A0)
#define TIMER0_PWMB_NORMAL_PORT_OPERATION \
    TCCR0A = (TCCR0A & ~_BV(COM0B1)) & ~_BV(COM0B0)
#define TIMER0_PWMB_CLEAR_ON_COMPARE_MATCH \
    TCCR0A = (TCCR0A | _BV(COM0B1)) & ~_BV(COM0B0)
#define TIMER0_PWMB_SET_ON_COMPARE_MATCH \
    TCCR0A = (TCCR0A | _BV(COM0B1)) | _BV(COM0B0)
// Timer/Counter0 Clock Prescaler
#define TIMER0_CLOCK_PRESCALER_0_STOP \
    TCCR0B = ((TCCR0B & ~_BV(CS02)) & ~_BV(CS01)) & ~_BV(CS00)
#define TIMER0_CLOCK_PRESCALER_1_START \
    TCCR0B = ((TCCR0B & ~_BV(CS02)) & ~_BV(CS01)) | _BV(CS00)
#define TIMER0_CLOCK_PRESCALER_8_START \
    TCCR0B = ((TCCR0B & ~_BV(CS02)) | _BV(CS01)) & ~_BV(CS00)
#define TIMER0_CLOCK_PRESCALER_64_START \
    TCCR0B = ((TCCR0B & ~_BV(CS02)) | _BV(CS01)) | _BV(CS00)
#define TIMER0_CLOCK_PRESCALER_256_START \
    TCCR0B = ((TCCR0B | _BV(CS02)) & ~_BV(CS01)) & ~_BV(CS00)
#define TIMER0_CLOCK_PRESCALER_1024_START \
    TCCR0B = ((TCCR0B | _BV(CS02)) & ~_BV(CS01)) | _BV(CS00)
#define TIMER0_EXTERNAL_CLOCK_FALLING_EDGE \
    TCCR0B = ((TCCR0B | _BV(CS02)) | _BV(CS01)) & ~_BV(CS00)
#define TIMER0_EXTERNAL_CLOCK_RISING_EDGE \
    TCCR0B = ((TCCR0B | _BV(CS02)) | _BV(CS01)) | _BV(CS00)
//
#define TIMER0_PWMA_PULSE_WIDTH(value) OCR0A = (value)
#define TIMER0_PWMB_PULSE_WIDTH(value) OCR0B = (value)
#define TIMER0_SET_VALUE(value) TCNT0 = (value)
#define TIMER0_COMPARE_MATCH_A_VALUE(value) OCR0A = (value)
// Timer/Counter0 Interrupt Flags
#define TIMER0_CLEAR_OVERFLOW_INTERRUPT_FLAG \
    TIFR0 |= _BV(TOV0)
#define TIMER0_CLEAR_COMPARE_MATCH_A_INTERRUPT_FLAG \
    TIFR0 |= _BV(OCF0A)
#define TIMER0_CLEAR_COMPARE_MATCH_B_INTERRUPT_FLAG \
    TIFR0 |= _BV(OCF0B)
// Timer/Counter0 Interrupts
#define TIMER0_OVERFLOW_INTERRUPT_ENABLE \
    TIMSK0 |= _BV(TOIE0)
#define TIMER0_OVERFLOW_INTERRUPT_DISABLE \
    TIMSK0 &= ~_BV(TOIE0)
#define TIMER0_COMPARE_MATCH_A_INTERRUPT_ENABLE \
    TIMSK0 |= _BV(OCIE0A)
#define TIMER0_COMPARE_MATCH_A_INTERRUPT_DISABLE \
    TIMSK0 &= ~_BV(OCIE0A)
#define TIMER0_COMPARE_MATCH_B_INTERRUPT_ENABLE \
    TIMSK0 |= _BV(OCIE0B)
#define TIMER0_COMPARE_MATCH_B_INTERRUPT_DISABLE \
    TIMSK0 &= ~_BV(OCIE0B)

/* Timer/Counter1 */
#define TIMER1_SET_VALUE(value) TCNT1 = (value)
#define TIMER1_VALUE TCNT1
#define TIMER1_RESET TCNT1 = 0x0000
// Modes of Operation
#define TIMER1_NORMAL_MODE \
    TCCR1B = (TCCR1B & ~_BV(WGM13)) & ~_BV(WGM12); \
    TCCR1A = (TCCR1A & ~_BV(WGM11)) & ~_BV(WGM10)
#define TIMER1_PHASE_CORRECT_PWM_8BIT \
    TCCR1B = (TCCR1B & ~_BV(WGM13)) & ~_BV(WGM12); \
    TCCR1A = (TCCR1A & ~_BV(WGM11)) | _BV(WGM10)
#define TIMER1_PHASE_CORRECT_PWM_9BIT \
    TCCR1B = (TCCR1B & ~_BV(WGM13)) & ~_BV(WGM12); \
    TCCR1A = (TCCR1A | _BV(WGM11)) & ~_BV(WGM10)
#define TIMER1_PHASE_CORRECT_PWM_10BIT \
    TCCR1B = (TCCR1B & ~_BV(WGM13)) & ~_BV(WGM12); \
    TCCR1A = (TCCR1A | _BV(WGM11)) | _BV(WGM10)
#define TIMER1_CLEAR_TIMER_ON_COMPARE_MATCH \
    TCCR1B = (TCCR1B & ~_BV(WGM13)) | _BV(WGM12); \
    TCCR1A = (TCCR1A & ~_BV(WGM11)) & ~_BV(WGM10)
#define TIMER1_FAST_PWM_8BIT \
    TCCR1B = (TCCR1B & ~_BV(WGM13)) | _BV(WGM12); \
    TCCR1A = (TCCR1A & ~_BV(WGM11)) | _BV(WGM10)
#define TIMER1_FAST_PWM_9BIT \
    TCCR1B = (TCCR1B & ~_BV(WGM13)) | _BV(WGM12); \
    TCCR1A = (TCCR1A | _BV(WGM11)) & ~_BV(WGM10)
#define TIMER1_FAST_PWM_10BIT \
    TCCR1B = (TCCR1B & ~_BV(WGM13)) | _BV(WGM12); \
    TCCR1A = (TCCR1A | _BV(WGM11)) | _BV(WGM10)
// Timer/Counter1 Clock Prescaler
#define TIMER1_CLOCK_PRESCALER_0_STOP \
    TCCR1B = ((TCCR1B & ~_BV(CS12)) & ~_BV(CS11)) & ~_BV(CS10)
#define TIMER1_CLOCK_PRESCALER_1_START \
    TCCR1B = ((TCCR1B & ~_BV(CS12)) & ~_BV(CS11)) | _BV(CS10)
#define TIMER1_CLOCK_PRESCALER_8_START \
    TCCR1B = ((TCCR1B & ~_BV(CS12)) | _BV(CS11)) & ~_BV(CS10)
#define TIMER1_CLOCK_PRESCALER_64_START \
    TCCR1B = ((TCCR1B & ~_BV(CS12)) | _BV(CS11)) | _BV(CS10)
#define TIMER1_CLOCK_PRESCALER_256_START \
    TCCR1B = ((TCCR1B | _BV(CS12)) & ~_BV(CS11)) & ~_BV(CS10)
#define TIMER1_CLOCK_PRESCALER_1024_START \
    TCCR1B = ((TCCR1B | _BV(CS12)) & ~_BV(CS11)) | _BV(CS10)
#define TIMER1_EXTERNAL_CLOCK_FALLING_EDGE \
    TCCR1B = ((TCCR1B | _BV(CS12)) | _BV(CS11)) & ~_BV(CS10)
#define TIMER1_EXTERNAL_CLOCK_RISING_EDGE \
    TCCR1B = ((TCCR1B | _BV(CS12)) | _BV(CS11)) | _BV(CS10)
// Timer/Counter1 Interrupt Flags
#define TIMER1_CLEAR_OVERFLOW_INTERRUPT_FLAG \
    TIFR1 |= _BV(TOV1)
#define TIMER1_CLEAR_COMPARE_MATCH_A_INTERRUPT_FLAG \
    TIFR1 |= _BV(OCF1A)
#define TIMER1_CLEAR_COMPARE_MATCH_B_INTERRUPT_FLAG \
    TIFR1 |= _BV(OCF1B)
#define TIMER1_CLEAR_INPUT_CAPTURE_INTERRUPT_FLAG \
    TIFR1 |= _BV(ICF1)
// Timer/Counter1 Interrupt Mask
#define TIMER1_OVERFLOW_INTERRUPT_ENABLE \
    TIMSK1 |= _BV(TOIE1)
#define TIMER1_OVERFLOW_INTERRUPT_DISABLE \
    TIMSK1 &= ~_BV(TOIE1)
#define TIMER1_COMPARE_MATCH_A_INTERRUPT_ENABLE \
    TIMSK1 |= _BV(OCIE1A)
#define TIMER1_COMPARE_MATCH_B_INTERRUPT_ENABLE \
    TIMSK1 |= _BV(OCIE1B)
#define TIMER1_INPUT_CAPTURE_INTERRUPT_ENABLE \
    TIMSK1 |= _BV(ICIE1)
#define TIMER1_INPUT_CAPTURE_INTERRUPT_DISABLE \
    TIMSK1 &= ~_BV(ICIE1)
// Timer/Counter1 Input Capture Noise Canceler
#define TIMER1_ACTIVATE_INPUT_CAPTURE_NOISE_CANCELER \
    TCCR1B |= _BV(ICNC1)
#define TIMER1_DEACTIVATE_INPUT_CAPTURE_NOISE_CANCELER \
    TCCR1B &= ~_BV(ICNC1)
// Timer/Counter1 Input Capture Edge Select
#define TIMER1_INPUT_CAPTURE_ON_RISING_EDGE \
    TCCR1B |= _BV(ICES1)
#define TIMER1_INPUT_CAPTURE_ON_FALLING_EDGE \
    TCCR1B &= ~_BV(ICES1)
#define TIMER1_TOGGLE_INPUT_CAPTURE_TRIGGER_EDGE \
    TCCR1B ^= _BV(ICES1)

/* Timer/Counter2 */
#define TIMER2_OVERFLOW \
    (TIFR2 & _BV(TOV2))
// Modes of Operation
#define TIMER2_NORMAL_MODE \
    TCCR2A &= ~_BV(WGM22); \
    TCCR2A = (TCCR2A & ~_BV(WGM21)) & ~_BV(WGM20)
#define TIMER2_PHASE_CORRECT_PWM_FULL_RANGE_MODE \
    TCCR2B &= ~_BV(WGM22); \
    TCCR2A = (TCCR2A & ~_BV(WGM21)) | _BV(WGM20)
#define TIMER2_CLEAR_TIMER_ON_COMPARE_MATCH_MODE \
    TCCR2B &= ~_BV(WGM22); \
    TCCR2A = (TCCR2A | _BV(WGM21)) & ~_BV(WGM20)
#define TIMER2_FAST_PWM_FULL_RANGE_MODE \
    TCCR2B &= ~_BV(WGM22); \
    TCCR2A = (TCCR2A | _BV(WGM21)) | _BV(WGM20)
#define TIMER2_PHASE_CORRECT_PWM_CONTROL_RANGE_MODE \
    TCCR2B |= _BV(WGM22); \
    TCCR2A = (TCCR2A & ~_BV(WGM21)) | _BV(WGM20)
#define TIMER2_FAST_PWM_CONTROL_RANGE_MODE \
    TCCR2B |= _BV(WGM22); \
    TCCR2A = (TCCR2A | _BV(WGM21)) | _BV(WGM20)
// Timer/Counter2 Clock Prescaler
#define TIMER2_CLOCK_PRESCALER_0_STOP \
    TCCR2B = ((TCCR2B & ~_BV(CS22)) & ~_BV(CS21)) & ~_BV(CS20)
#define TIMER2_CLOCK_PRESCALER_1_START \
    TCCR2B = ((TCCR2B & ~_BV(CS22)) & ~_BV(CS21)) | _BV(CS20)
#define TIMER2_CLOCK_PRESCALER_8_START \
    TCCR2B = ((TCCR2B & ~_BV(CS22)) | _BV(CS21)) & ~_BV(CS20)
#define TIMER2_CLOCK_PRESCALER_32_START \
    TCCR2B = ((TCCR2B & ~_BV(CS22)) | _BV(CS21)) | _BV(CS20)
#define TIMER2_CLOCK_PRESCALER_64_START \
    TCCR2B = ((TCCR2B | _BV(CS22)) & ~_BV(CS21)) & ~_BV(CS20)
#define TIMER2_CLOCK_PRESCALER_128_START \
    TCCR2B = ((TCCR2B | _BV(CS22)) & ~_BV(CS21)) | _BV(CS20)
#define TIMER2_CLOCK_PRESCALER_256_START \
    TCCR2B = ((TCCR2B | _BV(CS22)) | _BV(CS21)) & ~_BV(CS20)
#define TIMER2_CLOCK_PRESCALER_1024_START \
    TCCR2B = ((TCCR2B | _BV(CS22)) | _BV(CS21)) | _BV(CS20)
// Timer/Counter2 Interrupt Flags
#define TIMER2_CLEAR_OVERFLOW_INTERRUPT_FLAG \
    TIFR2 |= _BV(TOV2)
#define TIMER2_CLEAR_COMPARE_MATCH_A_INTERRUPT_FLAG \
    TIFR2 |= _BV(OCF2A)
#define TIMER2_CLEAR_COMPARE_MATCH_B_INTERRUPT_FLAG \
    TIFR2 |= _BV(OCF2B)
// Timer/Counter2 Interrupts
#define TIMER2_OVERFLOW_INTERRUPT_ENABLE \
    TIMSK2 |= _BV(TOIE2)
#define TIMER2_COMPARE_MATCH_A_INTERRUPT_ENABLE \
    TIMSK2 |= _BV(OCIE2A)
#define TIMER2_COMPARE_MATCH_B_INTERRUPT_ENABLE \
    TIMSK2 |= _BV(OCIE2B)
// Timer/Counter2 Asynchronous
#define TIMER2_ASYNCHRONOUS_CLOCK_INPUT \
    ASSR |= _BV(AS2)
//
#define TIMER2_VALUE TCNT2
#define TIMER2_RESET TCNT2 = 0x00
#define TIMER2_ASYNC_UPDATE (ASSR & _BV(TCN2UB))
#define TIMER2_PRESCALER_ASYNC_UPDATE (ASSR & _BV(TCR2BUB))

/* Analog Comparator */
#define ANALOG_COMPARATOR_ENABLE \
    ACSR &= ~_BV(ACD)
#define ANALOG_COMPARATOR_DISABLE \
    ACSR |= _BV(ACD)
#define ANALOG_COMPARATOR_INPUT_CAPTURE_ENABLE \
    ACSR |= _BV(ACIC)
#define ANALOG_COMPARATOR_INPUT_CAPTURE_DISABLE \
    ACSR &= ~_BV(ACIC)
#define ANALOG_COMPARATOR_BANDGAP_ENABLE \
    ACSR |= _BV(ACBG)
#define ANALOG_COMPARATOR_BANDGAP_DISABLE \
    ACSR &= ~_BV(ACBG)
#define ANALOG_COMPARATOR_CLEAR_INTERRUPT_FLAG \
    ACSR |= _BV(ACI)
#define ANALOG_COMPARATOR_INTERRUPT_ENABLE \
    ACSR |= _BV(ACIE)
#define ANALOG_COMPARATOR_INTERRUPT_DISABLE \
    ACSR &= ~_BV(ACIE)
#define ANALOG_COMPARATOR_INTERRUPT_ON_TOGGLE \
    ACSR = (ACSR & ~_BV(ACIS1)) & ~_BV(ACIS0)
#define ANALOG_COMPARATOR_INTERRUPT_ON_FALLING_EDGE \
    ACSR = (ACSR | _BV(ACIS1)) & ~_BV(ACIS0)
#define ANALOG_COMPARATOR_INTERRUPT_ON_RISING_EDGE \
    ACSR = (ACSR | _BV(ACIS1)) | _BV(ACIS0)
#define ANALOG_COMPARATOR_AIN0_DIGITAL_INPUT_DISABLE \
    DIDR |= _BV(AIN0D)
#define ANALOG_COMPARATOR_AIN0_DIGITAL_INPUT_ENABLE \
    DIDR &= ~_BV(AIN0D)
#define ANALOG_COMPARATOR_AIN1_DIGITAL_INPUT_DISABLE \
    DIDR |= _BV(AIN1D)
#define ANALOG_COMPARATOR_AIN1_DIGITAL_INPUT_ENABLE \
    DIDR &= ~_BV(AIN1D)

/*  Analog to Digital Converter */
// ADC Voltage Reference
#define ADC_VOLTAGE_REFERENCE_AREF \
    ADMUX = (ADMUX & ~_BV(REFS1)) & ~_BV(REFS0)
#define ADC_VOLTAGE_REFERENCE_AVCC \
    ADMUX = (ADMUX & ~_BV(REFS1)) | _BV(REFS0)
#define ADC_VOLTAGE_REFERENCE_INTERNAL \
    ADMUX = (ADMUX | _BV(REFS1)) | _BV(REFS0)
// ADC Adjust Result
#define ADC_RIGHT_ADJUST_RESULT \
    ADMUX &= ~_BV(ADLAR)
#define ADC_LEFT_ADJUST_RESULT \
    ADMUX |= _BV(ADLAR)
// ADC  Channel Selection
#define ADC_SELECT_CHANNEL_0 \
    ADMUX = (((ADMUX & ~_BV(MUX3)) & ~_BV(MUX2)) & ~_BV(MUX1)) & ~_BV(MUX0)
#define ADC_SELECT_CHANNEL_1 \
    ADMUX = (((ADMUX & ~_BV(MUX3)) & ~_BV(MUX2)) & ~_BV(MUX1)) | _BV(MUX0)
#define ADC_SELECT_CHANNEL_2 \
    ADMUX = (((ADMUX & ~_BV(MUX3)) & ~_BV(MUX2)) | _BV(MUX1)) & ~_BV(MUX0)
#define ADC_SELECT_CHANNEL_3 \
    ADMUX = (((ADMUX & ~_BV(MUX3)) & ~_BV(MUX2)) | _BV(MUX1)) | _BV(MUX0)
#define ADC_SELECT_CHANNEL_4 \
    ADMUX = (((ADMUX & ~_BV(MUX3)) | _BV(MUX2)) & ~_BV(MUX1)) & ~_BV(MUX0)
#define ADC_SELECT_CHANNEL_5 \
    ADMUX = (((ADMUX & ~_BV(MUX3)) | _BV(MUX2)) & ~_BV(MUX1)) | _BV(MUX0)
#define ADC_SELECT_CHANNEL_6 \
    ADMUX = (((ADMUX & ~_BV(MUX3)) | _BV(MUX2)) | _BV(MUX1)) & ~_BV(MUX0)
#define ADC_SELECT_CHANNEL_7 \
    ADMUX = (((ADMUX & ~_BV(MUX3)) | _BV(MUX2)) | _BV(MUX1)) | _BV(MUX0)
#define ADC_SELECT_TEMPERATURE_SENSOR \
    ADMUX = (((ADMUX | _BV(MUX3)) & ~_BV(MUX2)) & ~_BV(MUX1)) & ~_BV(MUX0)
#define ADC_SELECT_INTERNAL_VOLTAGE_REFERENCE \
    ADMUX = (((ADMUX | _BV(MUX3)) | _BV(MUX2)) | _BV(MUX1)) & ~_BV(MUX0)
#define ADC_SELECT_GROUND \
    ADMUX = (((ADMUX | _BV(MUX3)) | _BV(MUX2)) | _BV(MUX1)) | _BV(MUX0)
// ADC Status
#define ADC_DISABLE \
    ADCSRA &= ~_BV(ADEN)
#define ADC_ENABLE \
    ADCSRA |= _BV(ADEN)
#define ADC_START_CONVERSION \
    ADCSRA |= _BV(ADSC)
#define ADC_AUTO_TRIGGER_DISABLE \
    ADCSRA &= ~_BV(ADATE)
#define ADC_AUTO_TRIGGER_ENABLE \
    ADCSRA |= _BV(ADATE)
#define ADC_CLEAR_INTERRUPT_FLAG \
    ADCSRA |= _BV(ADIF)
#define ADC_INTERRUPT_ENABLE \
    ADCSRA |= _BV(ADIE)
#define ADC_FREE_RUNNING_MODE \
    ADCSRB = ((ADCSRB & ~_BV(ADTS2)) & ~_BV(ADTS1)) & ~_BV(ADTS0)
#define ADC_AUTO_TRIGGER_ANALOG_COMPARATOR \
    ADCSRB = ((ADCSRB & ~_BV(ADTS2)) & ~_BV(ADTS1)) | _BV(ADTS0)
#define ADC_AUTO_TRIGGER_EXTERNAL_INTERRUPT_0 \
    ADCSRB = ((ADCSRB & ~_BV(ADTS2)) | _BV(ADTS1)) & ~_BV(ADTS0)
#define ADC_AUTO_TRIGGER_TIMER0_COMPARE_MATCH_A \
    ADCSRB = ((ADCSRB & ~_BV(ADTS2)) | _BV(ADTS1)) | _BV(ADTS0)
#define ADC_AUTO_TRIGGER_TIMER0_OVERFLOW \
    ADCSRB = ((ADCSRB | _BV(ADTS2)) & ~_BV(ADTS1)) & ~_BV(ADTS0)
#define ADC_AUTO_TRIGGER_TIMER1_COMPARE_MATCH_B \
    ADCSRB = ((ADCSRB | _BV(ADTS2)) & ~_BV(ADTS1)) | _BV(ADTS0)
#define ADC_AUTO_TRIGGER_TIMER1_OVERFLOW \
    ADCSRB = ((ADCSRB | _BV(ADTS2)) | _BV(ADTS1)) & ~_BV(ADTS0)
#define ADC_AUTO_TRIGGER_TIMER1_CAPTURE_EVENT \
    ADCSRB = ((ADCSRB | _BV(ADTS2)) | _BV(ADTS1)) | _BV(ADTS0)
// ADC Prescaler
#define ADC_PRESCALER_2 \
    ADCSRA = ((ADCSRA & ~_BV(ADPS2)) & ~_BV(ADPS1)) | _BV(ADPS0)
#define ADC_PRESCALER_4 \
    ADCSRA = ((ADCSRA & ~_BV(ADPS2)) | _BV(ADPS1)) & ~_BV(ADPS0)
#define ADC_PRESCALER_8 \
    ADCSRA = ((ADCSRA & ~_BV(ADPS2)) | _BV(ADPS1)) | _BV(ADPS0)
#define ADC_PRESCALER_16 \
    ADCSRA = ((ADCSRA | _BV(ADPS2)) & ~_BV(ADPS1)) & ~_BV(ADPS0)
#define ADC_PRESCALER_32 \
    ADCSRA = ((ADCSRA | _BV(ADPS2)) & ~_BV(ADPS1)) | _BV(ADPS0)
#define ADC_PRESCALER_64 \
    ADCSRA = ((ADCSRA | _BV(ADPS2)) | _BV(ADPS1)) & ~_BV(ADPS0)
#define ADC_PRESCALER_128 \
    ADCSRA = ((ADCSRA | _BV(ADPS2)) | _BV(ADPS1)) | _BV(ADPS0)
// ADC Digital Input
#define ADC_DIGITAL_INPUT_ADC0_DISABLE \
    DIDR0 &= ~_BV(ADC0D)
#define ADC_DIGITAL_INPUT_ADC1_DISABLE \
    DIDR0 &= ~_BV(ADC1D)
#define ADC_DIGITAL_INPUT_ADC2_DISABLE \
    DIDR0 &= ~_BV(ADC2D)
#define ADC_DIGITAL_INPUT_ADC3_DISABLE \
    DIDR0 &= ~_BV(ADC3D)
#define ADC_DIGITAL_INPUT_ADC4_DISABLE \
    DIDR0 &= ~_BV(ADC4D)
#define ADC_DIGITAL_INPUT_ADC5_DISABLE \
    DIDR0 &= ~_BV(ADC5D)

/* External Interrupt */
#define EXTERNAL_INTERRUPT_0_REQUEST_ON_LOW_LEVEL \
    EICRA = (EICRA & ~_BV(ISC01)) & ~_BV(ISC00)
#define EXTERNAL_INTERRUPT_0_REQUEST_ON_TOGGLE \
    EICRA = (EICRA & ~_BV(ISC01)) | _BV(ISC00)
#define EXTERNAL_INTERRUPT_0_REQUEST_ON_FALLING_EDGE \
    EICRA = (EICRA | _BV(ISC01)) & ~_BV(ISC00)
#define EXTERNAL_INTERRUPT_0_REQUEST_ON_RISING_EDGE \
    EICRA = (EICRA | _BV(ISC01)) | _BV(ISC00)
#define EXTERNAL_INTERRUPT_1_REQUEST_ON_LOW_LEVEL \
    EICRA = (EICRA & ~_BV(ISC11)) & ~_BV(ISC10)
#define EXTERNAL_INTERRUPT_1_REQUEST_ON_TOGGLE \
    EICRA = (EICRA & ~_BV(ISC11)) | _BV(ISC10)
#define EXTERNAL_INTERRUPT_1_REQUEST_ON_FALLING_EDGE \
    EICRA = (EICRA | _BV(ISC11)) & ~_BV(ISC10)
#define EXTERNAL_INTERRUPT_1_REQUEST_ON_RISING_EDGE \
    EICRA = (EICRA | _BV(ISC11)) | _BV(ISC10)
// External Interrupt Flags
#define EXTERNAL_INTERRUPT_0_CLEAR_INTERRUPT_FLAG \
    EIFR |= _BV(INTF0)
#define EXTERNAL_INTERRUPT_1_CLEAR_INTERRUPT_FLAG \
    EIFR |= _BV(INTF1)
// External Interrupt Mask
#define EXTERNAL_INTERRUPT_REQUEST_0_ENABLE \
    EIMSK |= _BV(INT0)
#define EXTERNAL_INTERRUPT_REQUEST_0_DISABLE \
    EIMSK &= ~_BV(INT0)
#define EXTERNAL_INTERRUPT_REQUEST_1_ENABLE \
    EIMSK |= _BV(INT1)
#define EXTERNAL_INTERRUPT_REQUEST_1_DISABLE \
    EIMSK &= ~_BV(INT1)

/*  USART */
#define USART_RECEIVER_ENABLE \
    UCSR0B |= _BV(RXEN0);
#define USART_RECEIVER_DISABLE \
    UCSR0B &= ~_BV(RXEN0);
#define USART_TRAMSMITTER_ENABLE \
    UCSR0B |= _BV(TXEN0);
#define USART_TRAMSMITTER_DISABLE \
    UCSR0B &= ~_BV(TXEN0);
#define USART_SEND_BYTE(value) \
    UDR0 = (value)
#define USART_FRAME_ERROR \
    UCSR0A & _BV(FE0)
// USART Baud Rate
#define USART_PRESCALE(value) \
    ((F_CPU / ((value) * 16UL)) - 1)
#define USART_BAUD_RATE_2400 \
    UBRR0L = USART_PRESCALE(2400); \
    UBRR0H = (USART_PRESCALE(2400)>>8)
#define USART_BAUD_RATE_4800 \
    UBRR0L = USART_PRESCALE(4800); \
    UBRR0H = (USART_PRESCALE(4800)>>8)
#define USART_BAUD_RATE_9600 \
    UBRR0L = USART_PRESCALE(9600); \
    UBRR0H = (USART_PRESCALE(9600)>>8)
#define USART_BAUD_RATE_14400 \
    UBRR0L = USART_PRESCALE(14400); \
    UBRR0H = (USART_PRESCALE(14400)>>8)
#define USART_BAUD_RATE_19200 \
    UBRR0L = USART_PRESCALE(19200); \
    UBRR0H = (USART_PRESCALE(19200)>>8)
#define USART_BAUD_RATE_28800 \
    UBRR0L = USART_PRESCALE(28800); \
    UBRR0H = (USART_PRESCALE(28800)>>8)
#define USART_BAUD_RATE_38400 \
    UBRR0L = USART_PRESCALE(38400); \
    UBRR0H = (USART_PRESCALE(38400)>>8)
#define USART_BAUD_RATE_57600 \
    UBRR0L = USART_PRESCALE(57600); \
    UBRR0H = (USART_PRESCALE(57600)>>8)
#define USART_BAUD_RATE_76800 \
    UBRR0L = USART_PRESCALE(76800); \
    UBRR0H = (USART_PRESCALE(76800)>>8)
#define USART_BAUD_RATE_115200 \
    UBRR0L = USART_PRESCALE(115200); \
    UBRR0H = (USART_PRESCALE(115200)>>8)
// USART Interrupts
#define USART_RECEIVE_COMPLETE_INTERRUPT_ENABLE \
    UCSR0B |= _BV(RXCIE0);
#define USART_RECEIVE_COMPLETE_INTERRUPT_DISABLE \
    UCSR0B &= ~_BV(RXCIE0);
#define USART_TRAMSMIT_COMPLETE_INTERRUPT_ENABLE \
    UCSR0B |= _BV(TXCIE0);
#define USART_TRAMSMIT_COMPLETE_INTERRUPT_DISABLE \
    UCSR0B &= ~_BV(TXCIE0);
#define USART_DATA_REGISTER_EMPTY_INTERRUPT_ENABLE \
    UCSR0B |= _BV(UDRIE0);
#define USART_DATA_REGISTER_EMPTY_INTERRUPT_DISABLE \
    UCSR0B &= ~_BV(UDRIE0);
// USART Data Frame Length
#define USART_5BIT_DATA_FRAME \
    UCSR0B &= ~_BV(UCSZ02); \
    UCSR0C = (UCSR0C & ~_BV(UCSZ01)) & ~_BV(UCSZ00)
#define USART_6BIT_DATA_FRAME \
    UCSR0B &= ~_BV(UCSZ02); \
    UCSR0C = (UCSR0C & ~_BV(UCSZ01)) | _BV(UCSZ00)
#define USART_7BIT_DATA_FRAME \
    UCSR0B &= ~_BV(UCSZ02); \
    UCSR0C = (UCSR0C | _BV(UCSZ01)) & ~_BV(UCSZ00)
#define USART_8BIT_DATA_FRAME \
    UCSR0B &= ~_BV(UCSZ02); \
    UCSR0C = (UCSR0C | _BV(UCSZ01)) | _BV(UCSZ00)
#define USART_9BIT_DATA_FRAME \
    UCSR0B |= _BV(UCSZ02); \
    UCSR0C = (UCSR0C | _BV(UCSZ01)) | _BV(UCSZ00)
// USART Stop bit
#define USART_ONE_STOP_BIT \
    UCSR0C &= ~_BV(USBS0)
#define USART_TWO_STOP_BITS \
    UCSR0C |= _BV(USBS0)
// USART Parity Mode
#define USART_NO_PARITY \
    UCSR0C = (UCSR0C & ~_BV(UPM01)) & ~_BV(UPM00)
#define USART_PARITY_EVEN \
    UCSR0C = (UCSR0C | _BV(UPM01)) & ~_BV(UPM00)
#define USART_PARITY_ODD \
    UCSR0C = (UCSR0C | _BV(UPM01)) | _BV(UPM00)

#define ENABLE_INTERRUPTS sei()

#endif // ATMEGA88PA_H_INCLUDED
