#include <avr/io.h>
#include "ir_nec.h"

/* NEC Protocol Decoding */

// microseconds
#define START_PULSE_MIN 8000
#define START_PULSE_MAX 9999
#define START_SPACE_MIN 4000
#define START_SPACE_MAX 5000
#define SHORT_MIN 280
#define SHORT_MAX 840
#define LONG_MIN 1130
#define LONG_MAX 2250

typedef enum {
    START_PULSE,
    START_SPACE,
    BIT_PULSE,
    BIT_SPACE
} NEC_STATE;

static uint32_t nec, data;
static uint8_t counter;
static NEC_STATE state = START_PULSE;

#define EMMIT_0 counter++
#define EMMIT_1 nec |= ((uint32_t)1<<counter); counter++

void NEC_Reset(void) {
    counter = 0;
    nec = 0;
    state = START_PULSE;
}

uint8_t NEC_Address(void) {
    return data;
}

uint8_t NEC_Command(void) {
    return data>>16;
}

uint8_t NEC_Decode(uint16_t delay) {
    uint8_t ret = 0;
    switch(state) {
        case BIT_PULSE:
            if(counter == 32) {
                data = nec;
                ret = 1;
                //NEC42 (42-bit version of NEC protocol)
                uint8_t address = nec; //convert 13-bit address to 8-bits
                nec >>= 10;
                nec &= 0xFFFFFF00;
                nec |= address;
                counter = 22;
            }
            state = BIT_SPACE;
            if((delay < SHORT_MIN) || (delay > SHORT_MAX)) {
                NEC_Reset();
            }
            return ret;
        case BIT_SPACE:
            state = BIT_PULSE;
            if((delay > SHORT_MIN) && (delay < SHORT_MAX)) {
                EMMIT_0;
            } else if((delay > LONG_MIN) && (delay < LONG_MAX)) {
                EMMIT_1;
            } else { NEC_Reset(); }
            return 0;
        case START_PULSE:
            state = START_SPACE;
            if((delay < START_PULSE_MIN) || (delay > START_PULSE_MAX)) {
                NEC_Reset();
            }
            return 0;
        case START_SPACE:
            state = BIT_PULSE;
            data = 0;
            if((delay < START_SPACE_MIN) || (delay > START_SPACE_MAX)) {
                NEC_Reset();
            }
            return 0;
    }
    return 0;
}
