#include <avr/io.h>
#include "ATmega88PA.h"
#include "usart.h"
#include "stub.h"
#include "onewire.h"

#define ONEWIRE_START_TRANSIMSSION USART_DATA_REGISTER_EMPTY_INTERRUPT_ENABLE
#define ONEWIRE_STOP_TRANSIMSSION USART_DATA_REGISTER_EMPTY_INTERRUPT_DISABLE
#define ONEWIRE_RESET 0xF0
#define ONEWIRE_DEVICE_NOT_PRESENT 0xF0
#define ONEWIRE_WRITE_1 0xFF
#define ONEWIRE_WRITE_0 0x00
#define ONEWIRE_READ_1 0xFF

typedef enum {
    IDLE,
    RESET_PULSE,
    DATA_PULSE
} ONEWIRE_STATE;

static ONEWIRE_STATE state = IDLE;

static uint8_t *data_buffer;
static uint8_t data_pointer = 0;
static uint8_t data_lenght = 0;
static uint8_t tx_buffer = 0;
static uint8_t bits_sent = 0;
static uint8_t rx_buffer = 0;
static uint8_t bits_received = 0;

void ONEWIRE_SendReset(void) {
    USART_BAUD_RATE_9600;
    USART_SendData(ONEWIRE_RESET);
    state = RESET_PULSE;
    bits_received = 0;
}

uint8_t ONEWIRE_SendData(uint8_t *buffer, uint8_t length) {
    if(state != IDLE) { return 0; }
    USART_BAUD_RATE_115200;
    state = DATA_PULSE;
    data_buffer = buffer;
    data_lenght = length;
    data_pointer = 0;
    bits_sent = 0;
    bits_received = 0;
    ONEWIRE_START_TRANSIMSSION;
    return 1;
}

static void ONEWIRE_DataRegisterEmpty(void) {
    if(bits_sent == 8) {
        bits_sent = 0;
        data_pointer++;
        if(data_pointer == data_lenght) {
            ONEWIRE_STOP_TRANSIMSSION;
            return;
        }
    }
    if(bits_sent == 0) {
        tx_buffer = data_buffer[data_pointer];
    }
    if(tx_buffer & 0x01) {
        USART_SendData(ONEWIRE_WRITE_1);
    } else {
        USART_SendData(ONEWIRE_WRITE_0);
    }
    tx_buffer >>= 1;
    bits_sent++;
}

static void ONEWIRE_DataReceived(uint8_t data) {
    if(USART_FRAME_ERROR) {
        ONEWIRE_ReceiveError();
        state = IDLE;
        return;
    }
    if(state == RESET_PULSE) {
        if(data == ONEWIRE_DEVICE_NOT_PRESENT) {
            ONEWIRE_DeviceNotPresent();
        } else {
            ONEWIRE_DevicePresent();
        }
        state = IDLE;
        return;
    }
    bits_received++;
    rx_buffer >>= 1;
    if(data == ONEWIRE_READ_1) {
        rx_buffer |= 0x80;
    }
    if(bits_received == 8) {
        bits_received = 0;
        data_buffer[data_pointer-1] = rx_buffer;
        if(data_pointer == data_lenght) {
            state = IDLE;
            ONEWIRE_ReceiveComplete();
        }
    }
}

void ONEWIRE_Init(void) {
    ONEWIRE_ReceiveComplete = STUB_void;
    ONEWIRE_ReceiveError = STUB_void;
    ONEWIRE_DevicePresent = STUB_void;
    ONEWIRE_DeviceNotPresent = STUB_void;
}

void ONEWIRE_Activate(void) {
    USART_Setup_8N1();
    USART_DataReceived = ONEWIRE_DataReceived;
    USART_DataRegisterEmpty = ONEWIRE_DataRegisterEmpty;
    USART_Activate();
}
